import numpy as np

import os
os.chdir(os.path.dirname(__file__))

from SizeDistributionCalc import SizeDistributionCalc
from Conc2GenRate import Conc2GenRate
from ParticleSettlingVelocity import ParticleSettlingVelocity
from ViralDecayRate import ViralDecayRate
from RunFatima import RunFatima
from Diameter2PFUs import Diameter2PFUs
from Data2ViralConc import Data2ViralConc
from InhalabilityCalc import InhalabilityCalc
from DepositionCalc import DepositionCalc
from ViralConc2InhalationConc import ViralConc2InhalationConc
from InhalationConc2DepositionConc import InhalationConc2DepositionConc
from ProbitFcn import ProbitFcn

##############################################################################################################################################################
# DEFINE GLOBAL PARAMETERS ###################################################################################################################################
##############################################################################################################################################################

# Define mechanism of emission
emissionMech = 'coughing'

# Particle density, measured in g/cm^3
particleDensity = 1

# Minute ventilation of the infectious person, measured in L/min
minuteVentilation_dz = 15

# Minute ventilation of the exposed person, measured in L/min
minuteVentilation_ndz = 10

# Air density, measured in g/cm^3
airDensity = 0.0012041

# Evaporation diameter ratio
evaporationDiameterRatio = 0.29

# Viral load at day 0, measured in log viral copies/mL
medianViralLoad_day0 = 7.19
viralLoadSD_day0 = 1.35

# Infectivity ratio
infectivityRatio = 1e2

# Super spreader emission factor
SuperSpreaderEmissionFactor = 25

# Wind speed, measured in m/s
windSpeed = 0

# Find upow75 from wind speed
upow_75 =  np.power(windSpeed,2.75)/np.power(4, 2.75)

# Find nasal fraction and oral fraction from minute ventilation
if minuteVentilation_ndz<=35.5:
    nasalFraction = 1
else:
    nasalFraction = (0.29*minuteVentilation_ndz + 9.9)/(0.98*minuteVentilation_ndz + 1.4)
oralFraction = 1 - nasalFraction

# Median infectious dose, measured in TCID50
ID50 = 1

# Probit slope, measured in log10
logslope = 1

##############################################################################################################################################################
# DEFINE PARAMETERS RELATING TO SCENARIO (MECHANISM OF EMISSION) #############################################################################################
##############################################################################################################################################################

# Time parameters
c_hrs2sec = 60*60 
hrs = 24*1.5
T = hrs*c_hrs2sec

# Room size parameters, volume in m3 and surface area in m2
roomHeight = 4 # in m
roomWidth = 5 # in m
roomLength = 5 # in m
V = 10
Afloor = 40
Awalls = 2*roomWidth*roomHeight + 2*roomLength*roomHeight
Aceiling = Afloor
Aother = 4

# Ventilation volumetric flow rate, measured in m3/s
Qr = 100/c_hrs2sec

# Relative humidity, measured in %
relHumidity = 50
# Temperature, measured in C
temp = 20
# Simulated sunlight, measured in W/m2
sunlight = 1

# Define particle generation duration
Gt1 = 0*c_hrs2sec
Gt2 = 1*c_hrs2sec

# Exposure, assumed to be constant over the time interval [Et1, Et2]
Et1 = 0*c_hrs2sec
Et2 = 24*c_hrs2sec
    
params = {'particleDensity': particleDensity, 'airDensity': airDensity, 
          'evaporationDiameterRatio': evaporationDiameterRatio,
          'medianViralLoad_day0': medianViralLoad_day0, 'viralLoadSD_day0': viralLoadSD_day0, 
          'minuteVentilation_dz': minuteVentilation_dz, 'minuteVentilation_ndz': minuteVentilation_ndz,
          'infectivityRatio': infectivityRatio, 'SuperSpreaderEmissionFactor': SuperSpreaderEmissionFactor, 
          'emissionMech': emissionMech, 'windSpeed': windSpeed, 'upow_75': upow_75,
          'nasalFraction': nasalFraction, 'oralFraction': oralFraction, 'ID50': ID50, 'logslope': logslope}

# Define initial particle diameters in microns, bins per particle size decade,
#   and number of diameters to be put in the array
initialD = 0.1
decade = 10
numDiams = 31

# Count median diameter (um), geometric standard deviation, and 
#   concentration (particles/mL) for bronchiolar, laryngeal, and oral regions
if emissionMech=='breathing':
    
    params['gsd_bronch'] = 1.3
    params['cmd_bronch'] = 1.6
    params['concNum_bronch'] = 0.069
    params['cmd_laryn'] = None
    params['gsd_laryn'] = None
    params['concNum_laryn'] = None
    params['cmd_oral'] = None
    params['gsd_oral'] = None
    params['concNum_oral'] = None

elif emissionMech=='speaking':
    
    params['gsd_bronch'] = 1.3
    params['cmd_bronch'] = 1.6
    params['concNum_bronch'] = 0.069
    params['cmd_laryn'] = 2.5
    params['gsd_laryn'] = 1.68
    params['concNum_laryn'] = 0.086
    params['cmd_oral'] = 145
    params['gsd_oral'] = 1.8
    params['concNum_oral'] = 0.001

elif emissionMech=='coughing':
    
    params['gsd_bronch'] = 1.25
    params['cmd_bronch'] = 1.6
    params['concNum_bronch'] = 0.087
    params['cmd_laryn'] = 1.7
    params['gsd_laryn'] = 1.68
    params['concNum_laryn'] = 0.13
    params['cmd_oral'] = 123
    params['gsd_oral'] = 1.84
    params['concNum_oral'] = 0.016

##############################################################################################################################################################
# RUN COVID END-TO-END MODEL #################################################################################################################################
##############################################################################################################################################################

# Get array of pre-evaporation diameters
diametersPreEvap = np.zeros([numDiams])
diametersPreEvap[0] = initialD
for i in range(1, numDiams):
    diametersPreEvap[i] = diametersPreEvap[i-1]*np.power(10, 1/decade)

params['d0'] = diametersPreEvap
 
# Allow for evaporation and re-compute the resulting diameters
diameters = np.multiply(diametersPreEvap, evaporationDiameterRatio)
numDiams = len(diameters)

# Get the initial particle concentrations from the pre-evaporation diameters
concentration = SizeDistributionCalc(diameters, params)

# Get particle generation rates in #/min (converted to kg/s in RunFatima code)
particleGenerationRates = Conc2GenRate(concentration, params)
 
Time = np.arange(start=0, stop=T+1, step=15)
numTimePoints = len(Time)
particleData = np.zeros([numDiams, numTimePoints])

# These vectors are for sotring values so we can look at them if needed - can remove later
vd_floor_vec = np.zeros(numDiams)
k_inf_vec = np.zeros(numDiams)

# Set up parameters for FaTIMa
paramsFatima = {'V': V, 'Afloor': Afloor, 'Awalls': Awalls, 'Aceiling': Aceiling, 
                'Aother': Aother, 'Gt1': Gt1, 'Gt2': Gt2, 'Qr': Qr}

# Run FaTIMa model for each diameter
for i in range(numDiams):
    
    currD = diameters[i]
    currG = particleGenerationRates[i]
    
    # Get particle settling velocity in cm/s and convert to m/s
    vd_floor = ParticleSettlingVelocity(particleDensity, currD)
    vd_floor = vd_floor/100
    vd_floor_vec[i] = vd_floor
    
    # Get viral decay rate in 1/min and convert to 1/s
    k_inf, _ = ViralDecayRate(relHumidity, temp, sunlight)
    k_inf = k_inf/60
    k_inf_vec[i] = k_inf
    
    # Add generation rate, vd_floor, and k_inf to parameter dictionary
    paramsFatima['G'] = currG
    paramsFatima['vd_floor'] = vd_floor
    paramsFatima['k_inf'] = k_inf
    
    # Get FaTIMa output and save to particle matrix
    tsData = RunFatima(T, currD, particleDensity, paramsFatima)
    particleData[i,:] = tsData   

# Get PFU scaling factors
PFUs_median, PFUs_95pct = Diameter2PFUs(diametersPreEvap, params)

# Get viral concentration matrix
viralConcMatrix_median, viralConcMatrix_95pct = Data2ViralConc(particleData, PFUs_median, PFUs_95pct, 
                                                               SuperSpreaderEmissionFactor)

# Get inhalabiltiy scaling factors
inhalationFactor_median = InhalabilityCalc(diameters, windSpeed, upow_75, 
                                           minuteVentilation_ndz, nasalFraction, oralFraction)
inhalationFactor_95pct = inhalationFactor_median

# Get inhaled concentration matrix
inhalationConcMatrix_median, inhalationConcMatrix_95pct = ViralConc2InhalationConc(viralConcMatrix_median, viralConcMatrix_95pct, 
                                                                                   inhalationFactor_median, inhalationFactor_95pct)

# Get deposition fractions for each region
# Get deposition fractions for each region
depFactorHead_median, depFactorTB_median, depFactorLung_median = DepositionCalc(diameters, params['minuteVentilation_ndz'], inhalationFactor_median)
depFactorHead_95pct = depFactorHead_median.copy()
depFactorTB_95pct = depFactorTB_median.copy()
depFactorLung_95pct = depFactorLung_median.copy()

miscScalar = minuteVentilation_ndz*0.001*60
hourStepSize = 15/3600

# Get deposition matrix for each region
depConcMatrixHead_median, depConcMatrixHead_95pct = InhalationConc2DepositionConc(inhalationConcMatrix_median, inhalationConcMatrix_95pct, 
                                                                                  depFactorHead_median, depFactorHead_95pct, miscScalar, hourStepSize)
depConcMatrixTB_median, depConcMatrixTB_95pct = InhalationConc2DepositionConc(inhalationConcMatrix_median, inhalationConcMatrix_95pct, 
                                                                              depFactorTB_median, depFactorTB_95pct,  miscScalar, hourStepSize)
depConcMatrixLung_median, depConcMatrixLung_95pct = InhalationConc2DepositionConc(inhalationConcMatrix_median, inhalationConcMatrix_95pct,
                                                                                  depFactorLung_median, depFactorLung_95pct, miscScalar, hourStepSize)

# Rescale Et1 and Et2 to 15 second intervals (gives column index)
Et1 = int(Et1/15)
Et2 = int(Et2/15)

# Get total dose
dose_median = np.sum(depConcMatrixHead_median[:,Et1:Et2]) + np.sum(depConcMatrixTB_median[:,Et1:Et2]) + np.sum(depConcMatrixLung_median[:,Et1:Et2])
dose_95pct = np.sum(depConcMatrixHead_95pct[:,Et1:Et2]) + np.sum(depConcMatrixTB_95pct[:,Et1:Et2]) + np.sum(depConcMatrixLung_95pct[:,Et1:Et2])

# Get probability of infection
pInfected_median = ProbitFcn(dose_median, ID50, logslope)
pInfected_95pct = ProbitFcn(dose_95pct, ID50, logslope)

# Print results to the console
print('\n' + emissionMech)
print('  pInfected_median: {}'.format(pInfected_median))
print('  pInfected_95pct: {}'.format(pInfected_95pct))

##############################################################################################################################################################
# SAVE RESULTS ###############################################################################################################################################
##############################################################################################################################################################

from datetime import date
import pandas as pd

os.chdir(os.path.dirname(__file__))

if not os.path.isdir('Results'):
    os.mkdir('Results')
    
os.chdir(os.path.dirname(__file__) + '/Results')

# Create folder where results will be saved
currDate = date.today().strftime('%Y-%m-%d')
fldrName =currDate + '_' + emissionMech
if not os.path.isdir(fldrName):
    os.mkdir(fldrName)

# Define base path & file name for results
filePath = fldrName + '/' + currDate + '_' + emissionMech

# (1) Create DataFrame for Excel sheet with inputs/parameters
paramsFatima.pop('G')
paramsFatima.pop('vd_floor')
params.pop('d0')
descList = ['emissionMech'] + list(params.keys()) + list(paramsFatima.keys())
valList = [emissionMech] + [vv for kk, vv in params.items()] + [vv for kk, vv in paramsFatima.items()]
inputDF = pd.DataFrame({'Input/Parameter': descList, 'Value': valList})

# (2) Create DataFrame for Excel sheet with vector data
data_dict = {'diametersPreEvap': diametersPreEvap, 'diameters': diameters, 'particleGenRates': particleGenerationRates, 
             'vd_floor': vd_floor_vec, 'PFUs_median': PFUs_median, 'PFUs_95pct': PFUs_95pct, 
             'inhalationFactor_median': inhalationFactor_median, 'inhalationFactor_95pct': inhalationFactor_95pct,
             'depFactorHead_median': depFactorHead_median, 'depFactorTB_median': depFactorTB_median, 
             'depFactorLung_median': depFactorLung_median}
vectorDF = pd.DataFrame(data_dict)

# (3) Create DataFrames for Excel sheets with matrix data
timePts = Time
diam_columnNames = ['d = ' + str(round(tempD,3)) for tempD in diameters]
def getDF(numpy_data):
    df1 = pd.DataFrame({'Time': timePts})
    df2 = pd.DataFrame(data=np.transpose(numpy_data), columns=diam_columnNames)
    frames = [df1, df2]
    df = pd.concat(frames, axis=1)
    return df   
particleDF = getDF(particleData)
viralDF = getDF(viralConcMatrix_median)
inhalationDF = getDF(inhalationConcMatrix_median)
headDepDF = getDF(depConcMatrixHead_median)
tbDepDF = getDF(depConcMatrixTB_median)
pDepDF = getDF(depConcMatrixLung_median)

# (4) Create DataFrame for Excel sheet with outputs (total deposited dose and probability of infection)
output_dict = {'Output': ['Total deposited dose', 'Probability of infection'], 'Value': [dose_median, pInfected_median]}
outputDF = pd.DataFrame({'Output': ['Total deposited dose', 'Probability of infection'], 
                         'Value': [dose_median, pInfected_median]})

# (5) Create dictionary with keys = the name of each Excel sheet and values = corresponding data
excelSheets = {'Inputs': inputDF, 'Vector Data': vectorDF, 
               'Particle Data (FaTIMa)': particleDF, 'Viral Concentration': viralDF, 'Inhalation': inhalationDF, 
               'Head Deposition': headDepDF, 'TB Deposition': tbDepDF, 'Lung Deposition': pDepDF}

# (6) Save to Excel file
fN = filePath + '_results.xlsx'
writer = pd.ExcelWriter(fN, engine='xlsxwriter')
for sheet, frame in excelSheets.items(): frame.to_excel(writer, sheet_name=sheet, index=False)
writer.save()
del writer

##############################################################################################################################################################
# PLOT RESULTS ###############################################################################################################################################
##############################################################################################################################################################

from matplotlib import cm
import matplotlib.pyplot as plt

os.chdir(os.path.dirname(__file__) + '/Results/' + fldrName)

if not os.path.isdir(os.path.dirname(__file__) + '/Results/' + fldrName + '/Figures'):
    os.mkdir('Figures')

os.chdir(os.path.dirname(__file__) + '/Results/' + fldrName + '/Figures')

# Create color map for plotting time series data
cmColors = [cm.jet(i) for i in np.linspace(0,1,numDiams)]

times = np.arange(0, T+1, step=15)

# Plot particle concentration over time
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    plt.plot(times/c_hrs2sec, particleData[i,:], color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Particle Concentration over Time', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Particle Concentration (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99)
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('particleData.jpg')
plt.show()

# Plot median viral concentration over time
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    plt.plot(times/c_hrs2sec, viralConcMatrix_median[i,:], color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Viral Concentration over Time', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Viral Concentration (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99)
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('viralConc_median.jpg')
plt.show()

# Plot median inhaled concentration over time
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    plt.plot(times/c_hrs2sec, inhalationConcMatrix_median[i,:], color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Inhaled Concentration over Time', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Inhaled Concentration (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99)
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('inhaledConc_median.jpg')
plt.show()

# Plot median deposited dose over time (head)
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    plt.plot(times/c_hrs2sec, depConcMatrixHead_median[i,:], color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Deposited Dose over Time (Head)', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Deposited Dose (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99, loc='lower right')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('depositedDoseHead_median.jpg')
plt.show()

# Plot median deposited dose over time (TB)
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    plt.plot(times/c_hrs2sec, depConcMatrixTB_median[i,:], color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Deposited Dose over Time (TB)', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Deposited Dose (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99, loc='lower right')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('depositedDoseTB_median.jpg')
plt.show()

# Plot median deposited dose over time (Pulmonary)
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    plt.plot(times/c_hrs2sec, depConcMatrixLung_median[i,:], color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Deposited Dose over Time (Pulmonary)', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Deposited Dose (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99, loc='lower right')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('depositedDosePulmonary_median.jpg')
plt.show()

# Plot total medain deposited dose over time
plt.figure(figsize=(15,10))
for i, D in enumerate(diameters):
    totalConc = depConcMatrixHead_median[i,:] + depConcMatrixTB_median[i,:] + depConcMatrixLung_median[i,:]
    plt.plot(times/c_hrs2sec, totalConc, color=cmColors[i], linewidth=2, label='d = ' + str(round(D,3)))
plt.title('Total Deposited Dose over Time', fontsize=24)
plt.xlabel('log [ Time (h) ]', fontsize=22)
plt.ylabel('log [ Deposited Dose (#/m$^3$) ]', fontsize=22)
plt.xticks(fontsize=16)
plt.yticks(fontsize=16)
plt.yscale('log')
plt.xscale('log')
plt.legend(ncol=4, fontsize=11, fancybox=False, framealpha=0.99, loc='lower right')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'Times New Roman'
plt.rcParams['mathtext.fontset'] = 'stix'
plt.savefig('depositedDoseTotal_median.jpg')
plt.show()

os.chdir(os.path.dirname(__file__))